
#include "first.h"

#include "std_opt.h"


#include "raster_util.h"
#include "clut.h"
#include <assert.h>
#include "rast_i.h"
#include "sbuf.h"
#include "sbuf_i.h"


using namespace raster;



#ifdef __count_block_dims
int block_4_count,block_8_count,block_16_count;
#endif


namespace raster
{
 int __fastcall sb_init();
 void __fastcall sb_reset_for_draw(int _x_res, int _y_res);


  int clip_x_min, clip_x_max, 
      clip_y_min, clip_y_max;
  
  int min_y_span = max_y_res,max_y_span=0;

};




// note: this starts at 0 for terminals to alloc 2
span_block * sb_stk_ptr; //&span_block_stk[0];
span_block * span_block_stk; //&span_block_stk[0];

span_block * transparent_sb_stk_ptr; //&span_block_stk[0];
span_block * transparent_span_block_stk; //&span_block_stk[0];


arc_trans_group arc_trans[max_y_res+1];
sb_arc *        transparent_arc_starters[max_y_res+1];



inline u32 get_rgb_color(float shade, const solid_surface_desc * desc )
{
  u32 shade_id = f_to_int_us( shade, desc->shade_bits );
  u32 max = 1 << desc->shade_bits;
  if( shade_id >= max )
    shade_id = max-1;
  
  u32 color = ((u16*)desc->clut)[shade_id];

  return  color | (color<<16);
}



static float clip_x_max_f;



poly_ll_link * alt_polys=0;
poly_ll_link * als_polys=0;
poly_ll_link * afs_polys=0;
poly_ll_link * plt_polys=0;
poly_ll_link * pls_polys=0;

sbt_poly * tp_polys;


static_buffer<poly_edge_stk_size>  poly_edge_stk;
static_buffer<512*1024>            rend_dat_stk;

static_buffer<transparent_poly_stk_size>  transparent_poly_stk;







                           












//double sb_shade_magic =  fshl_hf( 52-ut_shade_bits-32-8);

//  ut_shade_magic = fshl_hf( 52-  p->surf->shade_bits -32-8);


double sb_shade_magic =  fshl_hf( 52-6-32 - 8);
//extern double sb_shade_magic =  fshl_hf( 52-6-32-8);

extern float  shade_trunc    =  fshl_hf( 52   );

 



















inline void insert_edge(sb_arc * arc, int y0, int y1)
{
 assert( y0 < y1 );
 assert( y0 >= clip_y_min );

 llist::add_head(arc_trans[y0].starters,arc);
 test_min_int(min_y_span,y0);
 
 if( y1 < clip_y_max  )
 {
  test_max_int(max_y_span,y1);

  arc->end_next = arc_trans[y1].enders; 
  arc_trans[y1].enders = arc;
 }
 else   // off the bottom
 {
  max_y_span = clip_y_max;
 }
}




void __fastcall raster::submit_edge(float x0,float y0, float x1, float y1,
                       poly * p0,
                       poly * p1) 
{  
    assert(p0 || p1 );
    assert_valid(p0);
    assert_valid(p1);
  //  assert( &p0 + 1 == &p1 ); 


    float rdx;
    {
     double tmp0 = y0 + d_to_fx16_magic; 
     double tmp1 = y1 + d_to_fx16_magic; 
     float _dx =  x1-x0;
     float _dy =  y1-y0 ;

     misc_qw_store[0] = tmp0;
     misc_qw_store[1] = tmp1;    
    

     // slope
     rdx = (_dx)/(_dy);
    }
    //
    // overlap with div
    //

    // round up
    int y_min = ((int&)misc_qw_store[0] +  FX1.i() - 1)>>16; //(iy0_5 + ((1<<__pos_frac_bits)-1)) >> __pos_frac_bits;
    int y_max = ((int&)misc_qw_store[1] +  FX1.i() - 1)>>16; //(iy1_5 + ((1<<__pos_frac_bits)-1)) >> __pos_frac_bits;
    
    int dy = y_max - y_min;
    int mask = (int)dy >> 31;

     y_min += (dy&mask);
     y_max -= (dy&mask);

     {
      u32 d_p = ((u32)p1 - (u32)p0)&mask;
      (u32&)p0 += d_p;
      (u32&)p1 -= d_p;
     }

    //
    // clipped?
    //
   test_max_int(y_min, clip_y_min);
   test_min_int(y_max, clip_y_max);


 

    // offset 
    {  
           // too far to the right?
     misc_qw_store[0] = (x0 + (float(y_min)  - y0)*rdx ) + d_to_fx16_magic;
     misc_qw_store[1] = rdx + d_to_fx16_magic;
    }

   
    u8 bad = ((u8( (int&)x0 >= (int&)clip_x_max_f )&
               u8( (int&)x1 >= (int&)clip_x_max_f) ) |
               u8(y_min >= y_max) );
    if( bad  )
	   { 
 	   return;
	   }


 

   sb_arc * arc;

   if( p0&&p1)
    {
      sb_arc_i * tsa = __alloc_edge(i);
      tsa->reset();
      arc= tsa;
      tsa->lp = p1;
      tsa->rp = p0;
    }
   else if( p1 )
     {
      sb_arc_r * tsa = __alloc_edge(r);
      tsa->reset();
      arc= tsa;
      tsa->lp = p1;
     }
   else
    {
      assert( p0);
      sb_arc_l * tsa = __alloc_edge(l);
      tsa->reset();
      arc= tsa;
      tsa->rp = p0;
     }

   

    


   arc->x  = /*(ix0_5 << (16-__pos_frac_bits) ) +*/                   // start x
             //(FX1.i() - (1 << (16-1-__pos_frac_bits)   ) ) +    // round_up
             (FX1.i() - 1) +    // round_up
             (int&)misc_qw_store[0];                             // pre-step


   arc->dx = (int&)misc_qw_store[1];



    insert_edge(arc,y_min,y_max);
    
}








void __fastcall  raster::submit_rect( const vert_3s & c, float width, float height,
                           const solid_surface_desc * sd  )
{

  int x0 = f_to_int_s( c[0] + 0.5f - width*0.5f );
  int x1 = f_to_int_s( c[0] + 0.5f + width*0.5f );
 
  if( x0 >= x1 ) 
    return;

  int y0 = f_to_int_s( c[1] + 0.5f - height*0.5f );
  int y1 = f_to_int_s( c[1] + 0.5f + height*0.5f );



  if( x1 <= clip_x_min )
   	 return;
  if( x0 >= clip_x_max )
   	 return;

  if( y1 <= clip_y_min )
   return;
  if( y0 >= clip_y_max )
   return;
  if( y0 < clip_y_min )
    y0 = clip_y_min;

  if( y0 >= y1 ) 
    return;

  poly * nply        = __alloc_poly();
  push_poly(afs_polys,nply);
  
  rend_data_f0 * rdat = alloc_rd(rdat);
  nply->reset(rdat);
  rdat->color = get_rgb_color(  c.shade, sd );
  rdat->dz[0] = 0.f;
  rdat->dz[1] = 0.f;
  rdat->z0 = c[2];


  sb_arc_l * la = __alloc_edge(l);
  la->reset();
  la->rp = nply;
  la->dx =0;
  la->x = x0*FX1.i();
  
  insert_edge(la, y0,y1);

  if(  x1 < clip_x_max )
  {
   sb_arc_r * ra = __alloc_edge(r);
   ra->reset();
   ra->lp = nply;
   ra->dx =0;
   ra->x = x1*FX1.i();

   insert_edge(ra, y0,y1);

  }
}





poly * __fastcall  raster::submit_als(const vert_3s & v0,
                          const vert_3s & v1,
                          const vert_3s & v2,
                          const solid_surface_desc * desc )
 {
  assert( v0.shade <= 1.0f );
  assert( v1.shade <= 1.0f );
  assert( v2.shade <= 1.0f );
  assert( v0.shade >= 0.0f );
  assert( v1.shade >= 0.0f );
  assert( v2.shade >= 0.0f );


  STD_P_SETUP()
  
  push_poly(als_polys,nply );

  rend_data_gour * rd = alloc_rd(rd);
  nply->reset(rd);

  rd->clut = desc->clut;

  SB_LB_TO_SCREEN_Z()
  
  SB_LB_TO_SCREEN_SH(rd->sh0,rd->dsh, v0.shade, v1.shade, v2.shade, 0 );

  return nply;
}



 poly * __fastcall  raster::submit_alt(const vert_3s & v0,
                          const vert_3s & v1,
                          const vert_3s & v2,
                          const  texture* tmap,
                          const float * uv0, const float * uv1, const float * uv2)
 {
//  assert( *(u32*)surf_desc->tmap  > 256 );

  STD_P_SETUP()

 
  push_poly(alt_polys,nply );

  rend_data_tex * rd = alloc_rd(rd);
  nply->reset(rd);
 
  // set texture
  rd->tmap = tmap;


  SB_LB_TO_SCREEN_Z()
  SB_LB_TO_SCREEN_SH(rd->sh0,rd->dsh, v0.shade, v1.shade, v2.shade, 1 );
  SB_LB_TO_SCREEN(rd->u0,rd->du,uv0[0],uv1[0],uv2[0]);
  SB_LB_TO_SCREEN(rd->v0,rd->dv,uv0[1],uv1[1],uv2[1]);
  

  return nply;
}









 inline int inc_wrap( int src, int max )
 {
	 return src +1 < max ? src + 1 : 0;
 }

inline int dec_wrap( int src, int max )
 {
	 return src-1 >= 0 ? src - 1 : max-1;
 }



extern "C" void __fastcall sb_fill_alt_tp(rend_data *rd, u32 * end_span);
extern "C" void __fastcall sb_fill_alt(rend_data *rd, u32 * end_span);

static void submit_tp_poly_edges(const vert_3 ** all_verts, 
							                          int num_verts,
                                 sbt_poly * nply)
{
   //
  // deal with edges
  //
  const int max_verts = 32;
  assert( num_verts < max_verts );

  int y_min = 1024*16, y_max=0;

  int ys[max_verts];
  for(int h=0;h<num_verts;++h)
  {
	  const vert_3 * vr = all_verts[h];
	  misc_qw_store[0] = vr->c[1] + d_to_fx16_magic; 
      // get y, rouned up
	  int y = ((int&)misc_qw_store[0] +  FX1.i() - 1)>>16; 
	  ys[h] = y;
	  test_min( y_min, y);
	  test_max( y_max, y);
  }
   //
   // clipped?
   //
  test_max_int(y_min, clip_y_min);
  test_min_int(y_max, clip_y_max);

  // no area covered?
  if( y_min >= y_max )
	   return;

  test_min_int(min_y_span,y_min);
  test_max_int(max_y_span,y_max);

  nply->end_y = y_max-1;
  // 
  // find first arc from start on left side
  //
  {
	  for(int iv1=0, iv0 = num_verts-1;;iv0=iv1,++iv1 )
	  {
		  assert( iv1 < num_verts );
		  if( (ys[iv0] <= y_min )&&(ys[iv1] > y_min) )
		  {
  		    break;
		  }
	  }
  
	  // write out left arcs
	  sbt_arc * out = (sbt_arc *)transparent_poly_stk.p;
	  nply->get_left_arc() = out;
 
	  int v0_y =y_min;
	  for(;;)
	  {
		    // v0, v1 are valid
		   int v1_y = __min(y_max,ys[iv1]);
		   int dy = v1_y - v0_y;
		   // edge covers ys?
		   if(dy > 0 )
		   {
		     const vert_3 & v0 = *all_verts[iv0];
		     const vert_3 & v1 = *all_verts[iv1];

		     float _dx =  v1.c[0]-v0.c[0];
		     float _dy =  v1.c[1]-v0.c[1];

		     // slope
		     float rdx = (_dx)/(_dy);
   
  		     // offset 
  		   misc_qw_store[0] = (v0.c[0] + (float(v0_y) - v0.c[1] )*rdx ) + d_to_fx16_magic;
		     misc_qw_store[1] = rdx + d_to_fx16_magic;
    
		     out->x  = (int&)misc_qw_store[0] + (FX1.i() - 1);
		     out->dx = (int&)misc_qw_store[1];
		     out->end_y = v1_y;
		     ++out;
		   }
		   // done?
		   if( v1_y == y_max )
		    break;
	    v0_y = v1_y;
	    iv0 = iv1;
		   iv1 = inc_wrap(iv1,num_verts);
	  }
 
   transparent_poly_stk.p =  out;
 }
  // 
  // find first arc from start on right side
  //
  {
	  
	  for(int iv0=0, iv1 = num_verts-1;;iv1=iv0,++iv0 )
	  {
		   assert( iv0 < num_verts );
		   if( (ys[iv0] <= y_min )&&(ys[iv1] > y_min) )
		   {
  		     break;
		   }
	  }
  
	  // write out right arcs
	  sbt_arc * out = (sbt_arc *)transparent_poly_stk.p;
	  nply->get_right_arc() = out;
 
	  //
	  // same as before except that we move backwards
	  //
	  int v0_y =y_min;
	  for(;;)
	  {
		    // v0, v1 are valid
		   int v1_y = __min(y_max,ys[iv1]);
		   int dy = v1_y - v0_y;
		   assert(dy >= 0);
		   // edge covers ys?
		   if(dy > 0 )
		   {
		     const vert_3 & v0 = *all_verts[iv0];
		     const vert_3 & v1 = *all_verts[iv1];

		     float _dx =  v1.c[0]-v0.c[0];
		     float _dy =  v1.c[1]-v0.c[1];

		     // slope
		     float rdx = (_dx)/(_dy);
   
  		     // offset 
  		     misc_qw_store[0] = (v0.c[0] + (float(v0_y) - v0.c[1] )*rdx ) + d_to_fx16_magic;
		     misc_qw_store[1] = rdx + d_to_fx16_magic;
    
		     out->x  = (int&)misc_qw_store[0] + (FX1.i() - 1);
		     out->dx = (int&)misc_qw_store[1];
		     out->end_y = v1_y;
		     ++out;
		   }
		   // done?
		   if( v1_y == y_max )
		    break;
     v0_y = v1_y;
		   iv0 = iv1;
		   iv1 = dec_wrap(iv1,num_verts);
	  }

   transparent_poly_stk.p =  out;

  }


  
  
  
  
  llist::add_head( transparent_poly_starts[y_min],nply );

}

void __fastcall raster::submit_alt_tp(const vert_3s & v0, // first tree verts
                             const vert_3s & v1,
                             const vert_3s & v2,
                             const texture * tmap,
                             const float  uv0[2],  // these are uv pairs for the first 3 verts
                             const float  uv1[2],
                             const float  uv2[2],
							 const vert_3 ** all_verts,  // all verts in poly.  Poly must be 
							                            // horizontally convex.
							 int num_verts)            // number of verts in poly

{
  STD_P_SETUP_TP()


 
  nply->sort_z = (v0[2]+v1[2]+v2[2])*(1.0f/3.0f);
  nply->fill_func = sb_fill_alt_tp;



  rend_data_tex * rd = d_alloc(rd,transparent_poly_stk);
  nply->reset(rd);
 
  // set texture
  rd->tmap = tmap;


  //
  // standard draw function
  //

  SB_LB_TO_SCREEN_Z()
  SB_LB_TO_SCREEN_SH(rd->sh0,rd->dsh, v0.shade, v1.shade, v2.shade, 1 );
  SB_LB_TO_SCREEN(rd->u0,rd->du,uv0[0],uv1[0],uv2[0]);
  SB_LB_TO_SCREEN(rd->v0,rd->dv,uv0[1],uv1[1],uv2[1]);
 
  assert( rd->z0 != 0.0f );


  submit_tp_poly_edges(all_verts, num_verts, nply);


//  llist::add_head( transparent_poly_starts[y_min],nply );
} 

void __fastcall raster::submit_aft_tp(const vert_3 & v0, // first tree verts
                             const vert_3 & v1,
                             const vert_3 & v2,
                             const float shade,
                             const texture * tmap,
                             const float  uv0[2],  // these are uv pairs for the first 3 verts
                             const float  uv1[2],
                             const float  uv2[2],
							 const vert_3 ** all_verts,  // all verts in poly.  Poly must be 
							                            // horizontally convex.
							 int num_verts)            // number of verts in poly

{
  STD_P_SETUP_TP()


 
  nply->sort_z = (v0[2]+v1[2]+v2[2])*(1.0f/3.0f);
  nply->fill_func = sb_fill_alt_tp;



  rend_data_tex * rd = d_alloc(rd,transparent_poly_stk);
  nply->reset(rd);
 
  // set texture
  rd->tmap = tmap;


  //
  // standard draw function
  //

  SB_LB_TO_SCREEN_Z()
  SB_LB_TO_SCREEN_SH(rd->sh0,rd->dsh, shade, shade, shade, 1 );
  SB_LB_TO_SCREEN(rd->u0,rd->du,uv0[0],uv1[0],uv2[0]);
  SB_LB_TO_SCREEN(rd->v0,rd->dv,uv0[1],uv1[1],uv2[1]);
 
  assert( rd->z0 != 0.0f );


  submit_tp_poly_edges(all_verts, num_verts, nply);


} 



//extern std_fill_func ** cur_sb_plt_draw_func;



void __fastcall raster::submit_plt_tp(const vert_3s & v0, // first tree verts
                             const vert_3s & v1,
                             const vert_3s & v2,
                             const texture * tmap,
                             const float  uv0[2],  // these are uv pairs for the first 3 verts
                             const float  uv1[2],
                             const float  uv2[2],
							 const vert_3 ** all_verts,  // all verts in poly.  Poly must be 
							                            // horizontally convex.
							 int num_verts)            // number of verts in poly

{
  STD_P_SETUP_TP()

  float z_av = (v0[2]+v1[2]+v2[2])*(1.0f/3.0f);
 
  nply->sort_z = z_av;
 // nply->fill_func =  cur_sb_plt_draw_func[1];
  


  rend_data_tex * rd = d_alloc(rd,transparent_poly_stk);
  nply->reset(rd);
 
  // set texture
  rd->tmap = tmap;


  //
  // standard draw function
  //

  SB_LB_TO_SCREEN_Z()
  
  nply->fill_func = get_persp_fill_func(__min(__min(v0[2],v1[2]),v2[2]), rd->dz,1);

  SB_LB_TO_SCREEN_P(rd->sh0,rd->dsh, v0.shade, v1.shade, v2.shade);
  SB_LB_TO_SCREEN_P(rd->u0,rd->du,uv0[0],uv1[0],uv2[0]);
  SB_LB_TO_SCREEN_P(rd->v0,rd->dv,uv0[1],uv1[1],uv2[1]);
 
  assert( rd->z0 != 0.0f );

  //
  // deal with edges
  //
  submit_tp_poly_edges(all_verts, num_verts, nply);


}


void __fastcall raster::submit_pft_tp(const vert_3 & v0, // first tree verts
                             const vert_3 & v1,
                             const vert_3 & v2,
                             const float shade,
                             const texture * tmap,
                             const float  uv0[2],  // these are uv pairs for the first 3 verts
                             const float  uv1[2],
                             const float  uv2[2],
							 const vert_3 ** all_verts,  // all verts in poly.  Poly must be 
							                            // horizontally convex.
							 int num_verts)            // number of verts in poly

{
  STD_P_SETUP_TP()

 
  float z_av = (v0[2]+v1[2]+v2[2])*(1.0f/3.0f);

  nply->sort_z = z_av;
//  nply->fill_func =  cur_sb_plt_draw_func[1];
  


  rend_data_tex * rd = d_alloc(rd,transparent_poly_stk);
  nply->reset(rd);
 
  // set texture
  rd->tmap = tmap;


  //
  // standard draw function
  //

  SB_LB_TO_SCREEN_Z()
  nply->fill_func = get_persp_fill_func(__min(__min(v0[2],v1[2]),v2[2]),rd->dz,1);
  SB_LB_TO_SCREEN_P(rd->sh0,rd->dsh, shade, shade, shade);
  SB_LB_TO_SCREEN_P(rd->u0,rd->du,uv0[0],uv1[0],uv2[0]);
  SB_LB_TO_SCREEN_P(rd->v0,rd->dv,uv0[1],uv1[1],uv2[1]);
 
  assert( rd->z0 != 0.0f );

  //
  // deal with edges
  //
  submit_tp_poly_edges(all_verts, num_verts, nply);
}





/*
   STD_P_SETUP()

 
  push_poly(plt_polys,nply );

  rend_data_tex_fn * rd = alloc_rd(rd);
  nply->reset(rd);
  rd->fill_func = cur_sb_plt_draw_func[0];
 
  // set texture
  rd->tmap = tmap;


  SB_LB_TO_SCREEN_Z()

  SB_LB_TO_SCREEN_P(rd->sh0,rd->dsh, v0.shade, v1.shade, v2.shade );
  SB_LB_TO_SCREEN_P(rd->u0,rd->du,uv0[0],uv1[0],uv2[0]);
  SB_LB_TO_SCREEN_P(rd->v0,rd->dv,uv0[1],uv1[1],uv2[1]);
  

  return nply;
*/












poly *  __fastcall raster::submit_aft(const vert_3 & v0,
                          const vert_3 & v1,
                          const vert_3 & v2,
                          float shade,
                          const  texture* tmap,
                          const float * uv0, const float * uv1, const float * uv2)
 {
//  assert( *(u32*)surf_desc->tmap  > 256 );

  STD_P_SETUP()

 
  push_poly(alt_polys,nply );
 
  rend_data_tex * rd = alloc_rd(rd);
  nply->reset(rd);
 
  // set texture
  rd->tmap = tmap;


  SB_LB_TO_SCREEN_Z()
  rd->sh0 = shade*linear_shade_scale + linear_shade_eps; 
  rd->dsh[0] = rd->dsh[1] = 0.f;
  SB_LB_TO_SCREEN(rd->u0,rd->du,uv0[0],uv1[0],uv2[0]);
  SB_LB_TO_SCREEN(rd->v0,rd->dv,uv0[1],uv1[1],uv2[1]);
  

  return nply;
}











 







poly *  __fastcall raster::submit_afs( const vert_3 & v0,
                            const vert_3 & v1,
                            const vert_3 & v2,
                            float shade ,
                           const solid_surface_desc * desc)
 {

  STD_P_SETUP()

  push_poly(afs_polys,nply );

  rend_data_f0 * rd = alloc_rd(rd);
  nply->reset(rd);

  SB_LB_TO_SCREEN_Z()

  // setup color

  rd->color =   get_rgb_color(shade, desc );


  return nply;
}


/*poly *  raster::submit_afs( const vert_3 & v0,
                            const vert_3 & v1,
                            const vert_3 & v2,
                            u32 color )
 {

 // no culling
  STD_P_SETUP()

  if( color != (1u<<16u) )
  {
   push_poly(afs_polys,nply );
  }

  rend_data_f0 * rd = alloc_rd(rd);
  nply->reset(rd);

  SB_LB_TO_SCREEN_Z()

  // setup color
  rd->color = color | (color << 16);

  return nply;
} */


















//
// this puts lines slightly in front of polys
// set to 1.0f for no adjustment
//
const float line_z_scale_eps = 257.0f/256.0f;

extern "C"
{
    double  __cdecl fabs(double);
    double  __cdecl sqrt(double);
};




void __fastcall raster::submit_line(const vert_3s & iv0, const vert_3s & iv1,
                       float thick0, float thick1,
                       const solid_surface_desc * desc )
{
 float v0[2],v1[2],v2[2],v3[2];
 
 float dx = (iv1[0]-iv0[0]);
 float dy = (iv1[1]-iv0[1]);
  
 float lg_ = 1.0f / sqrt( sqr(dx) + sqr(dy) );

 // normalize
 dx *= lg_;
 dy *= lg_;

 float tang_x0 =   dx* thick0*0.5f ;
 float tang_y0 =   dy* thick0*0.5f ;
 float tang_x1 =   dx* thick1*0.5f ;
 float tang_y1 =   dy* thick1*0.5f ;

 v0[0] = iv0[0] - tang_y0;
 v0[1] = iv0[1] + tang_x0;

 v1[0] = iv0[0] + tang_y0;
 v1[1] = iv0[1] - tang_x0;


 v2[0] = iv1[0] + tang_y1;
 v2[1] = iv1[1] - tang_x1;

 v3[0] = iv1[0] - tang_y1;
 v3[1] = iv1[1] + tang_x1;

 poly * nply = __alloc_poly();

 float stx  = (iv0[0]);
 float sty  = (iv0[1]);
 rend_data * rd;

 if( iv0.shade == iv1.shade  )
 {
  push_poly(afs_polys,nply );

  rend_data_f0 * _rd = alloc_rd(_rd);
  nply->reset(_rd);
  rd = _rd;
 
    // setup color
  _rd->color = get_rgb_color(iv0.shade, desc );

 }
 else
 {
  // linear shaded
  push_poly(als_polys,nply );

  rend_data_gour * _rd = alloc_rd( _rd );

  nply->reset(_rd);

  _rd->clut = desc->clut;
 
  float sh0  = linear_shade_eps + iv0.shade*linear_shade_scale ;
  float sh1  = linear_shade_eps + iv1.shade*linear_shade_scale ;

  float dsh = (sh1 - sh0);
 
  _rd->dsh[0] = (dx*lg_)*dsh; 
  _rd->dsh[1] = (dy*lg_)*dsh; 


  _rd->sh0   =  sh0 - _rd->dsh[0]*stx
                   - _rd->dsh[1]*sty;
  
  rd = _rd;
 

 }



 float z0  =  iv0[2]*line_z_scale_eps;
 float z1  =  iv1[2]*line_z_scale_eps;


 float dz = (z1 - z0);
 
 rd->dz[0] = (dx*lg_)*dz; 
 rd->dz[1] = (dy*lg_)*dz; 


 rd->z0  =  z0 - rd->dz[0]*stx
               - rd->dz[1]*sty;

		submit_edge( v0[0],v0[1],v1[0],v1[1],0,nply);
		submit_edge( v1[0],v1[1],v2[0],v2[1],0,nply);
		submit_edge( v2[0],v2[1],v3[0],v3[1],0,nply);
		submit_edge( v3[0],v3[1],v0[0],v0[1],0,nply);
}








#define __winapi(__rv) extern "C" __rv __stdcall

 //__winapi(int)    VirtualProtect(void*,u32,u32,unsigned long*);
 //__winapi(void*)  VirtualAlloc(void*aadr,u32 size,u32 type,u32 access_prot);
 //__winapi(u32)    GetLastError();
 //__winapi(int)    VirtualLock(void*,u32);


#define PAGE_EXECUTE_READWRITE 0x40     
#define PAGE_READWRITE         0x04     
#define MEM_COMMIT           0x1000     
#define MEM_RESERVE          0x2000     

static void __cdecl print_time_results();

extern void sweep_init();


extern "C" void __fastcall init_draw_sb(int is_565);



span_buffer_type span_buffer;


extern "C" int draw_sb_asm_start_0;
extern "C" int draw_sb_asm_end_0;
extern "C" int draw_sb_asm_start_1;
extern "C" int draw_sb_asm_end_1;





// used in asm routines
double misc_qw_store[10];




int __fastcall raster::sb_init()
{
  span_block_stk = (span_block*)VirtualAlloc(0, max_span_blocks*sizeof(span_block),MEM_COMMIT , PAGE_READWRITE);
  sb_stk_ptr = span_block_stk;


  transparent_span_block_stk = 
       (span_block*)VirtualAlloc(0, max_transparent_span_blocks*sizeof(span_block),MEM_COMMIT , PAGE_READWRITE);
  transparent_sb_stk_ptr = transparent_span_block_stk;
 




  sweep_init();


  unsigned long old =0;

 
  u32 asm_start  = (u32)__min(&draw_sb_asm_start_0,&draw_sb_asm_start_1);
  u32 asm_end   = (u32)__max(&draw_sb_asm_end_0,&draw_sb_asm_end_1);
  
  int size = asm_end - asm_start;
 
  if(!VirtualProtect( (void *)asm_start,size ,PAGE_EXECUTE_READWRITE,&old) )
   {
    return -1;
   }

 

/*  for(int x=0;x<max_x_res;++x)
   {
    int_to_flt[x] = float(x);
   }*/
 return 0;
}






extern void make_spans( int x0, int x1,u32 y);








//const int span_off_x = 0;
//const int span_off_y = 0;


void __fastcall raster::sb_reset_for_draw(int _x_res, int _y_res)
{
 //
 // save current float control word
 //
//  __asm fstcw math_cw_double;

 
 assert( _y_res  <= max_y_res );


/*  for(int y=0;y<_y_res;++y, (int&)buffer += buf_pitch)
  {
    frame_buf_starts[y] = (u16*)buffer;
  }*/

 clip_x_min =0;
 clip_y_min =0;

 clip_x_max = _x_res ;
 clip_x_max_f = float(clip_x_max);
 clip_y_max = _y_res;
 // x_res = _x_res;
// y_res = _y_res;

  // reset other stuff
  // conservative


// u32 span_tok = span_off_x | ( (_x_res-span_off_x)<<10);
 u32 span_tok ( (_x_res)<<10);
 
 //
 // reset span buffer to not drawn
 //
 for(int y=0;y<_y_res;++y)
  {
   span_buffer[y][0] = span_tok;
   span_buffer[y][1] = 0;
  }

}



